import csv

from django.core.management.base import BaseCommand

from wagtail_content_audit.query import PageSearchQuerySet
from wagtail_content_audit.utils import get_page_models_and_fields


class Command(BaseCommand):
    help = (
        "Search for a string or regular expression through page fields."
        "Pass a list in the form of app_name.page_type.field for each page "
        "type and field you want to report on. "
        "By default the report will include all page types and all text-based "
        "fields on the page."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-p",
            "--pagetype",
            action="append",
            help=(
                "Specify the page type(s) and field to check."
                "This should be given in the form app_name.page_type.field "
                "to include a page type in the given app with the given field."
                "For example, v1.BrowsePage.content."
            ),
        )
        parser.add_argument(
            "-s",
            "--search",
            required=True,
            help=(
                "The search string to match. This can be a regular expression."
            ),
        )

    def handle(self, *args, **options):
        search_string = options["search"]
        pagetypes = options["pagetype"]

        search_qs = PageSearchQuerySet().filter(search=search_string)
        if pagetypes is not None:
            for page_model, field_name in get_page_models_and_fields(
                pagetypes
            ):
                search_qs = search_qs.filter(
                    page_model=page_model, field=field_name
                )

        writer = csv.writer(self.stdout)
        writer.writerow(
            (
                "Page ID",
                "Page Type",
                "Page Title",
                "Page URL",
                "Field",
                "Field Type",
                "Stream Field Path",
                "Block Type",
                "Result Path",
                "Stream Field Matches",
            )
        )
        for result in search_qs.all():
            writer.writerow(
                (
                    result.page.id,
                    result.page_model.__name__,
                    result.page.title,
                    result.page.url,
                    result.field_name,
                    result.field_type,
                    ".".join(result.stream_field_path),
                    result.block_type,
                    ".".join(result.result_path),
                    *result.matches,
                )
            )



from io import StringIO

from django.core.management import call_command
from django.test import TestCase


class BlockUsageCommandTestCase(TestCase):
    fixtures = ["wagtail_content_audit_testapp_fixture.json"]

    def test_usage(self):
        output = StringIO()
        call_command("block_usage", stdout=output)
        self.assertIn("streamfield_with_block,block", output.getvalue())
        self.assertIn("streamfield_with_list,list.item", output.getvalue())
        self.assertIn("streamfield_with_table,table.text", output.getvalue())

    def test_usage_with_page_type_and_field(self):
        output = StringIO()
        call_command(
            "block_usage",
            "-p",
            "testapp.SearchTestPage.streamfield_with_block",
            "-p",
            "testapp.SearchTestPage.streamfield_with_list",
            stdout=output,
        )
        self.assertIn("streamfield_with_block,block", output.getvalue())
        self.assertIn("streamfield_with_list,list.item", output.getvalue())
        self.assertNotIn(
            "streamfield_with_table,table.text", output.getvalue()
        )



from io import StringIO

from django.core.management import call_command
from django.test import TestCase


class PageSearchCommandTestCase(TestCase):
    fixtures = ["wagtail_content_audit_testapp_fixture.json"]

    def test_search_streamfield(self):
        output = StringIO()
        call_command("page_search", "-s", "Test", stdout=output)
        self.assertIn(",struct.givenname", output.getvalue())
        self.assertIn(",0.struct.givenname", output.getvalue())

    def test_search_regular_field(self):
        output = StringIO()
        call_command("page_search", "-s", "text", stdout=output)
        self.assertIn(
            "text",
            output.getvalue(),
        )

    def test_search_with_page_type_and_field(self):
        output = StringIO()
        call_command(
            "page_search",
            "-s",
            "Test",
            "-p",
            "testapp.SearchTestPage.streamfield_with_block",
            "-p",
            "testapp.SearchTestPage.streamfield_with_list",
            stdout=output,
        )
        self.assertIn("0.block", output.getvalue())
        self.assertIn("0.list.item.0", output.getvalue())
        self.assertNotIn("0.struct.givenname", output.getvalue())



from django.test import TestCase

from wagtail_content_audit.query.blockusage import (
    BlockUsageQuerySet,
    traverse_streamblock,
    traverse_streamvalue,
)
from wagtail_content_audit.tests.testapp.models import SearchTestPage


class BlockUsageTestCase(TestCase):
    fixtures = ["wagtail_content_audit_testapp_fixture.json"]

    def setUp(self):
        self.page_one = SearchTestPage.objects.get(id=3)
        self.page_two = SearchTestPage.objects.get(id=4)

    def test_traverse_streamblock_streamblock(self):
        streamfield = SearchTestPage._meta.get_field("streamfield_with_block")
        results = list(
            traverse_streamblock(
                SearchTestPage,
                next(iter(streamfield.stream_block.child_blocks.items()))[1],
            )
        )
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].path, "block")

    def test_traverse_streamblock_listblock(self):
        streamfield = SearchTestPage._meta.get_field("streamfield_with_list")
        results = list(
            traverse_streamblock(
                SearchTestPage,
                next(iter(streamfield.stream_block.child_blocks.items()))[1],
            )
        )
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].path, "list")
        self.assertEqual(results[1].path, "list.item")

    def test_traverse_streamblock_tableblock(self):
        streamfield = SearchTestPage._meta.get_field("streamfield_with_table")
        results = list(
            traverse_streamblock(
                SearchTestPage,
                next(iter(streamfield.stream_block.child_blocks.items()))[1],
            )
        )
        self.assertEqual(len(results), 3)
        self.assertEqual(results[0].path, "table")
        self.assertEqual(results[1].path, "table.text")
        self.assertEqual(results[2].path, "table.numeric")

    def test_traverse_streamvalue_boundblock(self):
        results = list(
            traverse_streamvalue(self.page_one.streamfield_with_block)
        )
        self.assertEqual(results, ["block"])

    def test_traverse_streamvalue_structvalue(self):
        results = traverse_streamvalue(self.page_one.streamfield_with_struct)
        self.assertEqual(
            list(results), ["struct", "struct.givenname", "struct.surname"]
        )

    def test_traverse_streamvalue_listvalue(self):
        results = traverse_streamvalue(self.page_one.streamfield_with_list)
        self.assertEqual(list(results), ["list", "list.item", "list.item"])

    def test_traverse_streamvalue_table(self):
        results = traverse_streamvalue(self.page_one.streamfield_with_table)
        self.assertEqual(
            list(results),
            [
                "table",
                "table.text",
                "table.numeric",
                "table.text",
                "table.numeric",
            ],
        )

    def test_blockusagequeryset_get_filtered_page_models_no_filter(self):
        queryset = BlockUsageQuerySet()
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 2)
        self.assertIn(SearchTestPage, page_models)

    def test_blockusagequeryset_get_filtered_page_models_with_filter(self):
        queryset = BlockUsageQuerySet().filter(page_model=SearchTestPage)
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 1)
        self.assertIn(SearchTestPage, page_models)

    def test_blockusagequeryset_get_filtered_page_models_with_filter_str(self):
        queryset = BlockUsageQuerySet().filter(page_model="SearchTestPage")
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 1)
        self.assertIn(SearchTestPage, page_models)

    def test_blockusagequeryset_get_filtered_page_models_filter_dne(self):
        queryset = BlockUsageQuerySet().filter(page_model="Foo")
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 0)

    def test_blockusagequeryset_get_filtered_streamfield_names_no_filter(self):
        queryset = BlockUsageQuerySet()
        field_names = list(
            queryset.get_filtered_streamfield_names(page_model=SearchTestPage)
        )
        self.assertEqual(len(field_names), 4)

    def test_blockusagequeryset_get_filtered_streamfield_names_with_filter(
        self,
    ):
        queryset = BlockUsageQuerySet().filter(field="streamfield_with_block")
        field_names = list(
            queryset.get_filtered_streamfield_names(page_model=SearchTestPage)
        )
        self.assertEqual(len(field_names), 1)

    def test_blockusagequeryset_get_filtered_streamfield_names_filter_dne(
        self,
    ):
        queryset = BlockUsageQuerySet().filter(field="nonexistent_field")
        field_names = list(
            queryset.get_filtered_streamfield_names(page_model=SearchTestPage)
        )
        self.assertEqual(len(field_names), 0)

    def test_blockusagequeryset_audit_blocks_for_page_model(self):
        queryset = BlockUsageQuerySet()
        results = queryset.audit_blocks_for_page_model(SearchTestPage)

        self.assertEqual(len(results), 4)
        self.assertIn("streamfield_with_block", results)
        self.assertIn("streamfield_with_list", results)
        self.assertIn("streamfield_with_table", results)
        self.assertIn("streamfield_with_struct", results)

        block = results["streamfield_with_block"]["block"]
        self.assertEqual(len(block.pages), 2)
        self.assertIn(self.page_one, block.pages)

    def test_blockusagequeryset_run_query(self):
        queryset = BlockUsageQuerySet()
        results = queryset
        self.assertEqual(len(results), 9)

        results = queryset[1:2]
        self.assertEqual(len(results), 1)



from django.db import models

from wagtail import blocks as blocks
from wagtail.admin.panels import FieldPanel
from wagtail.contrib.typed_table_block.blocks import TypedTableBlock
from wagtail.fields import StreamField
from wagtail.models import Page


class SearchTestPage(Page):
    streamfield_with_block = StreamField(
        [
            ("block", blocks.CharBlock()),
        ],
        use_json_field=True,
    )

    streamfield_with_list = StreamField(
        [
            ("list", blocks.ListBlock(blocks.CharBlock())),
        ],
        use_json_field=True,
    )

    streamfield_with_struct = StreamField(
        [
            (
                "struct",
                blocks.StructBlock(
                    [
                        ("givenname", blocks.CharBlock()),
                        ("surname", blocks.CharBlock()),
                    ]
                ),
            ),
        ],
        use_json_field=True,
    )

    streamfield_with_table = StreamField(
        [
            (
                "table",
                TypedTableBlock(
                    [
                        ("text", blocks.CharBlock()),
                        ("numeric", blocks.FloatBlock()),
                    ]
                ),
            )
        ],
        use_json_field=True,
    )

    text = models.TextField(blank=True)

    content_panels = Page.content_panels + [
        FieldPanel("streamfield_with_block"),
        FieldPanel("streamfield_with_list"),
        FieldPanel("streamfield_with_struct"),
        FieldPanel("streamfield_with_table"),
        FieldPanel("text"),
    ]



# Generated by Django 3.2.23 on 2024-02-29 14:50

import django.db.models.deletion
from django.db import migrations, models

import wagtail.blocks
import wagtail.contrib.typed_table_block.blocks
import wagtail.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0089_log_entry_data_json_null_to_object'),
    ]

    operations = [
        migrations.CreateModel(
            name='SearchTestPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.page')),
                ('streamfield_with_block', wagtail.fields.StreamField([('block', wagtail.blocks.CharBlock())], use_json_field=True)),
                ('streamfield_with_list', wagtail.fields.StreamField([('list', wagtail.blocks.ListBlock(wagtail.blocks.CharBlock()))], use_json_field=True)),
                ('streamfield_with_struct', wagtail.fields.StreamField([('struct', wagtail.blocks.StructBlock([('givenname', wagtail.blocks.CharBlock()), ('surname', wagtail.blocks.CharBlock())]))], use_json_field=True)),
                ('streamfield_with_table', wagtail.fields.StreamField([('table', wagtail.contrib.typed_table_block.blocks.TypedTableBlock([('text', wagtail.blocks.CharBlock()), ('numeric', wagtail.blocks.FloatBlock())]))], use_json_field=True)),
                ('text', models.TextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
    ]



import csv

from django.core.management.base import BaseCommand

from wagtail_content_audit.query import BlockUsageQuerySet
from wagtail_content_audit.utils import get_page_models_and_fields


class Command(BaseCommand):
    help = (
        "Search for a string or regular expression through page fields."
        "Pass a list in the form of app_name.page_type.field for each page "
        "type and field you want to report on. "
        "By default the report will include all page types and all text-based "
        "fields on the page."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "-p",
            "--pagetype",
            action="append",
            help=(
                "Specify the page type(s) and field to check."
                "This should be given in the form app_name.page_type.field "
                "to include a page type in the given app with the given field."
                "For example, v1.BrowsePage.content."
            ),
        )

    def handle(self, *args, **options):
        pagetypes = options["pagetype"]

        audited_blocks_qs = BlockUsageQuerySet()

        if pagetypes is not None:
            for page_model, field_name in get_page_models_and_fields(
                pagetypes
            ):
                audited_blocks_qs = audited_blocks_qs.filter(
                    page_model=page_model, field=field_name
                )

        writer = csv.writer(self.stdout)
        writer.writerow(
            [
                "Page Type",
                "Field",
                "Path",
                "Block",
                "Occurrences",
                "Pages",
                "Live",
                "In Default Site",
            ]
        )
        for audited_block in audited_blocks_qs.all():
            writer.writerow(
                [
                    audited_block.page_model,
                    audited_block.field,
                    audited_block.path,
                    audited_block.block,
                    audited_block.total_occurrences,
                    audited_block.pages_count,
                    audited_block.pages_live_count,
                    audited_block.pages_in_default_site_count,
                ]
            )



import itertools
import logging
import re
from dataclasses import dataclass

from django.core.exceptions import FieldError

from wagtail.blocks import BoundBlock, StreamValue, StructValue
from wagtail.blocks.list_block import ListValue
from wagtail.contrib.typed_table_block.blocks import TypedTable
from wagtail.models import Page, Site, get_page_models

from queryish import Queryish

from wagtail_content_audit.utils import dotted_name


logger = logging.getLogger(__name__)


@dataclass
class PageMatch:
    page_model: type
    page: Page
    field_name: str
    field_type: str
    stream_field_path: list
    block_type: type
    result_path: list
    matches: list


def search_blocks(pattern, value, path=None):
    if path is None:
        path = []

    if isinstance(value, BoundBlock):
        local_path = path + [value.block]
        yield from search_blocks(pattern, value.value, path=local_path)

    elif isinstance(value, StructValue):
        local_path = path
        for child in value.bound_blocks.values():
            yield from search_blocks(pattern, child, path=path)

    elif isinstance(value, ListValue):
        for index, child in enumerate(value.bound_blocks):
            local_path = path + ["item", index]
            yield from search_blocks(pattern, child, path=local_path)

    elif isinstance(value, StreamValue):
        for index, child in enumerate(value):
            local_path = path + [index]
            yield from search_blocks(pattern, child, path=local_path)

    elif isinstance(value, TypedTable):
        for row_index, row in enumerate(value.rows):
            local_path = path + [row_index]
            for column_index, child in enumerate(row):
                local_path = local_path + [column_index]
                yield from search_blocks(pattern, child, path=local_path)

    else:
        matches = pattern.findall(str(value))
        if len(matches) > 0:
            yield path, matches


class PageSearchQuerySet(Queryish):
    def get_filtered_page_models(self):
        global_page_models = get_page_models()
        filters = [val for key, val in self.filters if key == "page_model"]
        filtered_page_models = [
            page_model
            for page_model in global_page_models
            if page_model in filters or page_model.__name__ in filters
        ]
        return filtered_page_models if len(filters) > 0 else global_page_models

    def get_filtered_field_names(self, page_model):
        all_fields = [field.name for field in page_model._meta.concrete_fields]
        filters = [val for key, val in self.filters if key == "field"]
        filtered_fields = [
            field_name for field_name in all_fields if field_name in filters
        ]
        return filtered_fields if len(filters) > 0 else all_fields

    def get_search_re(self):
        search_str = next(
            (val for key, val in self.filters if key == "search"), ""
        )
        search_re = re.compile(search_str)
        return search_re

    def prepare_pattern_for_json(self, pattern):
        return pattern.replace('"', r'\\"')

    def get_matches_for_page_field(self, page_model, field_name, page):
        search_re = self.get_search_re()
        field = page_model._meta.get_field(field_name)
        field_type = dotted_name(field.__class__)
        field_value = getattr(page, field_name)

        page_match = PageMatch(
            page_model=page_model,
            page=page,
            field_name=field_name,
            field_type=field_type,
            stream_field_path=[],
            block_type=None,
            result_path=[],
            matches=[],
        )

        if isinstance(field_value, StreamValue):
            # If this field is a StreamField, dive into it to get paths
            # and matches
            for streamfield_path, matches in search_blocks(
                search_re, field_value
            ):
                # Construct a specific path (with list indexes) and then a
                # general path that follow the conventions of Wagtail's
                # StreamField migration pathing.
                formatted_path = [
                    (
                        (p.name if hasattr(p, "name") else p)
                        if p is not None
                        else "item"
                    )
                    for p in streamfield_path
                ]
                page_match.block_type = dotted_name(
                    streamfield_path[-1].__class__
                )
                page_match.result_path = [str(p) for p in formatted_path]
                page_match.stream_field_path = [
                    p for p in formatted_path if not isinstance(p, int)
                ]
                page_match.matches = matches
                yield page_match

        else:
            matches = search_re.findall(str(field_value))
            page_match.matches = matches
            yield page_match

    def get_matches_for_page_model_field(self, page_model, field_name):
        search_re = self.get_search_re()

        # Get the default site
        site = Site.objects.get(is_default_site=True)

        # Search for live pages in the default site
        queryset = page_model.objects.live().in_site(site)

        # Try to do in-database regular expression-matching of the search
        # string.
        try:
            queryset = queryset.filter(
                **{
                    f"{field_name}__iregex": self.prepare_pattern_for_json(
                        search_re.pattern
                    )
                }
            )
        except FieldError:
            logger.info(
                f"Cannot search {dotted_name(page_model)}.{field_name}."
            )
            return

        queryset = queryset.exact_type(page_model)
        for page in queryset:
            yield from self.get_matches_for_page_field(
                page_model, field_name, page
            )

    def run_query(self):
        search_matches = []

        for page_model in self.get_filtered_page_models():
            for field_name in self.get_filtered_field_names(page_model):
                search_matches = itertools.chain(
                    search_matches,
                    self.get_matches_for_page_model_field(
                        page_model, field_name
                    ),
                )

        # Slice based on queryset slicing offset/limit
        return itertools.islice(
            search_matches,
            self.offset,
            self.offset + self.limit if self.limit else None,
            self.limit,
        )



from wagtail.models import get_page_models


dotted_name = lambda cls: ".".join((cls.__module__, cls.__qualname__))


def get_page_models_and_fields(pagetypes=None):
    page_models = get_page_models()

    for page_model in page_models:
        for field in page_model._meta.concrete_fields:
            pagetype_str = (
                f"{page_model._meta.app_label}."
                f"{page_model._meta.object_name}."
                f"{field.name}"
            )
            if pagetypes is None or pagetype_str in pagetypes:
                yield page_model, field.name



import re

from django.test import TestCase

from wagtail_content_audit.query.pagesearch import (
    PageSearchQuerySet,
    search_blocks,
)
from wagtail_content_audit.tests.testapp.models import SearchTestPage


class PageSearchTestCase(TestCase):
    fixtures = ["wagtail_content_audit_testapp_fixture.json"]

    def setUp(self):
        self.test_page = SearchTestPage.objects.get(id=3)
        self.notest_page = SearchTestPage.objects.get(id=4)

    def test_search_blocks_boundblock(self):
        pattern = re.compile("Test")

        result = list(
            search_blocks(pattern, self.test_page.streamfield_with_block)
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0][0]), 2)
        self.assertEqual(result[0][1], ["Test"])

        notest_result = list(
            search_blocks(pattern, self.notest_page.streamfield_with_block)
        )
        self.assertEqual(len(notest_result), 0)

    def test_search_blocks_structvalue(self):
        pattern = re.compile("Test")

        result = list(
            search_blocks(pattern, self.test_page.streamfield_with_struct)
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0][0]), 3)
        self.assertEqual(result[0][1], ["Test"])

        notest_result = list(
            search_blocks(pattern, self.notest_page.streamfield_with_struct)
        )
        self.assertEqual(len(notest_result), 0)

    def test_search_blocks_listvalue(self):
        pattern = re.compile("Test")

        result = list(
            search_blocks(pattern, self.test_page.streamfield_with_list)
        )
        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0][0]), 5)
        self.assertEqual(result[0][1], ["Test"])

        notest_result = list(
            search_blocks(pattern, self.notest_page.streamfield_with_list)
        )
        self.assertEqual(len(notest_result), 0)

    def test_search_blocks_typedtable(self):
        pattern = re.compile("Test")

        result = list(
            search_blocks(pattern, self.test_page.streamfield_with_table)
        )
        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0][0]), 5)
        self.assertEqual(result[0][1], ["Test"])

        notest_result = list(
            search_blocks(pattern, self.notest_page.streamfield_with_table)
        )
        self.assertEqual(len(notest_result), 0)

    def test_pagesearchqueryset_get_filtered_page_models_no_filter(self):
        queryset = PageSearchQuerySet()
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 2)
        self.assertIn(SearchTestPage, page_models)

    def test_pagesearchqueryset_get_filtered_page_models_with_filter(self):
        queryset = PageSearchQuerySet().filter(page_model=SearchTestPage)
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 1)
        self.assertIn(SearchTestPage, page_models)

    def test_pagesearchqueryset_get_filtered_page_models_with_filter_str(self):
        queryset = PageSearchQuerySet().filter(page_model="SearchTestPage")
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 1)
        self.assertIn(SearchTestPage, page_models)

    def test_pagesearchqueryset_get_filtered_page_models_filter_dne(self):
        queryset = PageSearchQuerySet().filter(page_model="Foo")
        page_models = queryset.get_filtered_page_models()
        self.assertEqual(len(page_models), 0)

    def test_pagesearchqueryset_get_filtered_field_names_no_filter(self):
        queryset = PageSearchQuerySet()
        field_names = list(
            queryset.get_filtered_field_names(page_model=SearchTestPage)
        )
        self.assertTrue(len(field_names) > 6)

    def test_pagesearchqueryset_get_filtered_field_names_with_filter(self):
        queryset = PageSearchQuerySet().filter(field="streamfield_with_block")
        field_names = list(
            queryset.get_filtered_field_names(page_model=SearchTestPage)
        )
        self.assertEqual(len(field_names), 1)

    def test_pagesearchqueryset_get_filtered_field_names_filter_dne(self):
        queryset = PageSearchQuerySet().filter(field="nonexistent_field")
        field_names = list(
            queryset.get_filtered_field_names(page_model=SearchTestPage)
        )
        self.assertEqual(len(field_names), 0)

    def test_pagesearchqueryset_get_search_re(self):
        queryset = PageSearchQuerySet().filter(search="Test")
        search_re = queryset.get_search_re()
        self.assertEqual(search_re.pattern, "Test")

    def test_pagesearchqueryset_prepare_pattern_for_json(self):
        queryset = PageSearchQuerySet()
        json_pattern = queryset.prepare_pattern_for_json(
            'pattern with "quotes"'
        )
        self.assertEqual(json_pattern, 'pattern with \\\\"quotes\\\\"')

    def test_pagesearchqueryset_get_matches_for_page_field_streamfield(self):
        queryset = PageSearchQuerySet().filter(search="Test")
        match = next(
            queryset.get_matches_for_page_field(
                SearchTestPage, "streamfield_with_block", self.test_page
            )
        )
        self.assertEqual(len(match.stream_field_path), 1)
        self.assertEqual(len(match.result_path), 2)
        self.assertEqual(len(match.matches), 1)

    def test_pagesearchqueryset_get_matches_for_page_field_nonstreamfield(
        self,
    ):
        queryset = PageSearchQuerySet().filter(search="Test")
        match = next(
            queryset.get_matches_for_page_field(
                SearchTestPage, "text", self.test_page
            )
        )
        self.assertEqual(len(match.stream_field_path), 0)
        self.assertEqual(len(match.result_path), 0)
        self.assertEqual(len(match.matches), 1)

    def test_pagesearchqueryset_get_matches_for_page_model_field(self):
        queryset = PageSearchQuerySet().filter(search="Test")
        match = next(
            queryset.get_matches_for_page_model_field(
                SearchTestPage,
                "text",
            )
        )
        self.assertEqual(len(match.stream_field_path), 0)
        self.assertEqual(len(match.result_path), 0)
        self.assertEqual(len(match.matches), 1)

    def test_pagesearchqueryset_get_matches_for_page_model_field_noiregex(
        self,
    ):
        queryset = PageSearchQuerySet().filter(search="Test")
        with (
            self.assertRaises(StopIteration),
            self.assertLogs(level="INFO") as logging_context,
        ):
            next(
                queryset.get_matches_for_page_model_field(
                    SearchTestPage,
                    "latest_revision",
                )
            )
        self.assertEqual(len(logging_context.output), 1)
        self.assertIn(
            (
                "Cannot search wagtail_content_audit.tests.testapp.models."
                "SearchTestPage.latest_revision."
            ),
            logging_context.output[0],
        )

    def test_pagesearchqueryset_run_query(self):
        queryset = PageSearchQuerySet().filter(search="Test")
        self.assertEqual(queryset.count(), 11)



from django.test import TestCase

from wagtail.models import Page

from wagtail_content_audit.tests.testapp.models import SearchTestPage
from wagtail_content_audit.utils import dotted_name, get_page_models_and_fields


class UtilsTestCase(TestCase):

    def test_dotted_name(self):
        name = dotted_name(SearchTestPage)
        self.assertEqual(
            name, "wagtail_content_audit.tests.testapp.models.SearchTestPage"
        )

    def test_get_page_models_and_fields(self):
        page_models_and_fields = list(get_page_models_and_fields())
        self.assertIn((Page, "title"), page_models_and_fields)
        self.assertIn(
            (SearchTestPage, "streamfield_with_block"), page_models_and_fields
        )

        pagetypes = ["testapp.SearchTestPage.streamfield_with_block"]
        page_models_and_fields = list(get_page_models_and_fields(pagetypes))
        self.assertNotIn((Page, "title"), page_models_and_fields)
        self.assertIn(
            (SearchTestPage, "streamfield_with_block"), page_models_and_fields
        )



from dataclasses import dataclass

from wagtail.blocks import (
    BoundBlock,
    ListBlock,
    StreamBlock,
    StreamValue,
    StructBlock,
    StructValue,
)
from wagtail.blocks.list_block import ListValue
from wagtail.contrib.typed_table_block.blocks import (
    TypedTable,
    TypedTableBlock,
)
from wagtail.models import Site, get_page_models

from queryish import Queryish

from wagtail_content_audit.utils import dotted_name


@dataclass
class AuditedBlock:
    page_model: type
    field: str
    path: str
    block: type
    pages: list
    total_occurrences: int = 0
    pages_count: int = 0
    pages_live_count: int = 0
    pages_in_default_site_count: int = 0


# Traverse a stream field and yield back each available block type
def traverse_streamblock(page_model, block, parent=None):
    """Walk model stream block objects to get initial AuditedBlocks"""
    block_name = block.name if block.name != "" else "item"
    audited_block = AuditedBlock(
        page_model=dotted_name(page_model),
        field=None,
        path=(parent + "." + block_name if parent is not None else block_name),
        block=dotted_name(block.__class__),
        pages=[],
    )

    yield audited_block

    # If this is a StreamBlock or StructBlock it'll have child blocks
    if isinstance(block, (StreamBlock, StructBlock, TypedTableBlock)):
        for child_block_field_name in block.child_blocks:
            yield from traverse_streamblock(
                page_model,
                block.child_blocks[child_block_field_name],
                parent=audited_block.path,
            )

    elif isinstance(block, ListBlock):
        yield from traverse_streamblock(
            page_model,
            block.child_block,
            parent=audited_block.path,
        )

    else:
        pass


# Traverse a stream field's value and yield back each block type in use
def traverse_streamvalue(value, parent=None):
    """Walk model stream value objects to get AuditedBlocks in-use"""

    # This is a block type
    if isinstance(value, BoundBlock):
        block_name = value.block.name if value.block.name != "" else "item"
        path = parent + "." + block_name if parent is not None else block_name
        yield path
        yield from traverse_streamvalue(value.value, parent=path)

    # This is a StructValue
    elif isinstance(value, StructValue):
        for child in value.bound_blocks.values():
            yield from traverse_streamvalue(child, parent=parent)

    elif isinstance(value, ListValue):
        for child in value.bound_blocks:
            yield from traverse_streamvalue(child, parent=parent)

    elif isinstance(value, TypedTable):
        for row_index, row in enumerate(value.rows):
            for column_index, child in enumerate(row):
                yield from traverse_streamvalue(child, parent=parent)

    # This is a sequence of blocks
    elif isinstance(value, StreamValue):
        for child in value:
            yield from traverse_streamvalue(child, parent=parent)


class BlockUsageQuerySet(Queryish):
    """Return a QuerySet-like object for querying block type usage"""

    def get_filtered_page_models(self):
        global_page_models = get_page_models()
        filters = [val for key, val in self.filters if key == "page_model"]
        filtered_page_models = [
            page_model
            for page_model in global_page_models
            if page_model in filters or page_model.__name__ in filters
        ]
        return filtered_page_models if len(filters) > 0 else global_page_models

    def get_filtered_streamfield_names(self, page_model):
        all_streamfields = page_model.get_streamfield_names()
        filters = [val for key, val in self.filters if key == "field"]
        filtered_fields = [
            field_name
            for field_name in all_streamfields
            if field_name in filters
        ]
        return filtered_fields if len(filters) > 0 else all_streamfields

    def audit_blocks_for_page_model(self, page_model):
        # Get the StreamFields on the page model
        streamfields = self.get_filtered_streamfield_names(page_model)

        # A dictionary to hold counts of each block on a page
        page_blocks = {}

        # First populate all available blocks
        for streamfield_name in streamfields:
            streamfield = page_model._meta.get_field(streamfield_name)
            page_blocks[streamfield_name] = {}

            # Traverse child blocks of this stream field (this avpathoids capturing
            # the containing stream block, so we can line-up with stream values
            # below)
            for child_block_name in streamfield.stream_block.child_blocks:
                for audited_block in traverse_streamblock(
                    page_model,
                    streamfield.stream_block.child_blocks[child_block_name],
                ):
                    audited_block.field = streamfield_name
                    page_blocks[streamfield_name][
                        audited_block.path
                    ] = audited_block

        # Get the default Wagtail site (this avoids the Trash)
        site = Site.objects.get(is_default_site=True)

        # Get a queryset for all pages of this type
        page_queryset = page_model.objects.exact_type(page_model)

        # Loop through the queryset, and traverse each streamfield
        for page in page_queryset:
            for streamfield_name in streamfields:
                streamfield_value = getattr(page, streamfield_name)
                streamfield_dict = page_blocks[streamfield_name]

                for block_path in traverse_streamvalue(streamfield_value):
                    # Get the AuditedBlock object for this path
                    audited_block = streamfield_dict[block_path]

                    audited_block.total_occurrences += 1

                    if page not in audited_block.pages:
                        audited_block.pages.append(page)

                        audited_block.pages_count += 1

                        if page.live:
                            audited_block.pages_live_count += 1

                        if page.is_descendant_of(site.root_page):
                            audited_block.pages_in_default_site_count += 1

        return page_blocks

    def run_query(self):
        audited_blocks = []

        for page_model in self.get_filtered_page_models():
            page_blocks = self.audit_blocks_for_page_model(page_model)

            # Flatten the dictionary of dictionaries of blocks to a list and
            # extend the audited block list
            audited_blocks.extend(
                block
                for streamfield, blocks in page_blocks.items()
                for block in blocks.values()
            )

        # self.ordering

        # Slice based on queryset slicing offset/limit
        return audited_blocks[
            self.offset : self.offset + self.limit if self.limit else None
        ]
